/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl.hibernate;

import net.sf.autodao.AutoDAO;
import org.springframework.beans.factory.annotation.AnnotatedBeanDefinition;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.BeanDefinitionHolder;
import org.springframework.beans.factory.support.AbstractBeanDefinition;
import org.springframework.beans.factory.support.BeanDefinitionBuilder;
import static org.springframework.beans.factory.support.BeanDefinitionBuilder.genericBeanDefinition;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.beans.factory.xml.BeanDefinitionParser;
import org.springframework.beans.factory.xml.ParserContext;
import org.springframework.beans.factory.xml.XmlReaderContext;
import org.springframework.context.annotation.ClassPathBeanDefinitionScanner;
import org.springframework.core.type.filter.AnnotationTypeFilter;
import static org.springframework.util.StringUtils.commaDelimitedListToStringArray;
import org.w3c.dom.Element;

/**
 * Processes <code>dao:hibernateScan</code> elements and builds Spring beans.
 * <p/>
 * Creates {@link HibernateDao} beans and wrap it with {@link HibernateDaoProxyFactoryBean} so that AutoDAO magic
 * works.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
public final class HibernateScanBeanDefinitionParser implements BeanDefinitionParser {
  private static final String BASE_PACKAGE_ATTRIBUTE = "base-package";
  private static final String RESOURCE_PATTERN_ATTRIBUTE = "resource-pattern";

  @Override
  public BeanDefinition parse(final Element element, final ParserContext parserContext) {
    final HibernateDaoScanner scanner = new HibernateDaoScanner(parserContext.getRegistry());
    final XmlReaderContext xmlReaderContext = parserContext.getReaderContext();
    if (element.hasAttribute(RESOURCE_PATTERN_ATTRIBUTE)) {
      scanner.setResourcePattern(element.getAttribute(RESOURCE_PATTERN_ATTRIBUTE));
    }
    scanner.setResourceLoader(xmlReaderContext.getResourceLoader());
    scanner.setBeanDefinitionDefaults(parserContext.getDelegate().getBeanDefinitionDefaults());
    scanner.setAutowireCandidatePatterns(parserContext.getDelegate().getAutowireCandidatePatterns());
    scanner.scan(
        element,
        commaDelimitedListToStringArray(element.getAttribute(BASE_PACKAGE_ATTRIBUTE)));
    return null;
  }

  private static class HibernateDaoScanner extends ClassPathBeanDefinitionScanner {

    public HibernateDaoScanner(final BeanDefinitionRegistry registry) {
      super(registry, true);
    }

    protected final void scan(final Element element, final String[] basePackages) {
      for (final String basePackage : basePackages) {
        for (final BeanDefinition candidate : findCandidateComponents(basePackage)) {
          final String daoInterface = candidate.getBeanClassName();
          final BeanDefinitionBuilder builder = genericBeanDefinition(HibernateDaoProxyFactoryBean.class);
          ParserUtils.applyDaoParams(builder, element, daoInterface, null);
          final AbstractBeanDefinition definition = builder.getBeanDefinition();
          final String beanName = "net.sf.autodao.AutoDAO[" + daoInterface + ']';
          postProcessBeanDefinition(definition, beanName);
          if (checkCandidate(beanName, definition)) {
            registerBeanDefinition(new BeanDefinitionHolder(definition, beanName), getRegistry());
          }
        }
      }
    }

    @Override
    protected void registerDefaultFilters() {
      addIncludeFilter(new AnnotationTypeFilter(AutoDAO.class, false));
    }

    @Override
    protected boolean isCandidateComponent(final AnnotatedBeanDefinition beanDefinition) {
      return beanDefinition.getMetadata().isInterface();
    }
  }
}
