/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl.hibernate;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import net.sf.autodao.Named;
import net.sf.autodao.QueryArgumentTransformer;
import net.sf.autodao.impl.SpecificationViolationException;
import net.sf.autodao.impl.Utils;
import org.hibernate.QueryParameterException;
import org.hibernate.engine.query.NamedParameterDescriptor;
import org.hibernate.engine.query.OrdinalParameterDescriptor;
import org.hibernate.engine.query.ParameterMetadata;
import org.hibernate.type.Type;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * Performs finder method arguments validation.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
final class ParametersChecker implements Utils.ParameterCallback {
  private final boolean singleFind;
  @NotNull
  private final Method method;
  @Nullable
  private final ParameterMetadata metadata;
  @NotNull
  private final Map<Class<?>, QueryArgumentTransformer> transformers;

  @NotNull
  private final Set<String> names = new HashSet<String>();
  private boolean limit;
  private boolean offset;
  private int indexed;

  public ParametersChecker(final boolean singleFind,
                           @NotNull final Method method,
                           @Nullable final ParameterMetadata metadata,
                           @NotNull final Map<Class<?>, QueryArgumentTransformer> transformers) {
    this.singleFind = singleFind;
    this.method = method;
    this.metadata = metadata;
    this.transformers = transformers;
  }

  @Override
  public void visit(final int index, @NotNull final Class<?> type, @NotNull final Annotation[] annotations) {
    boolean processed = false;
    if (Utils.getLimit(annotations) != null) {
      limitCheck(type);
      processed = true;
    }
    if (Utils.getOffset(annotations) != null) {
      offsetCheck(type);
      processed = true;
    }
    final Named named = Utils.getNamed(annotations);
    if (processed) {
      if (named != null) {
        throw new SpecificationViolationException("@Named cannot be present on @Limit/@Offset param", method);
      }
    } else {
      if (named == null) {
        if (!names.isEmpty()) {
          throw new SpecificationViolationException("You're not allowed to mix @Named and indexed params", method);
        }
        if (metadata != null) {
          final int paramIndex = index + 1
              - (limit ? 1 : 0)
              - (offset ? 1 : 0);
          final OrdinalParameterDescriptor descriptor;
          try {
            descriptor = metadata.getOrdinalParameterDescriptor(paramIndex);
          } catch (final IndexOutOfBoundsException e) {
            throw new SpecificationViolationException("Query doesn't have enough indexed parameters", method, e);
          }
          final Class<?> expectedType = descriptor.getExpectedType().getReturnedClass();
          final Class<?> actualType = getActualParameterType(type, false);
          if (!expectedType.isAssignableFrom(actualType)) {
            final String msg = String.format("Indexed parameter '%s' of type %s cannot be assigned to query parameter of type %s", paramIndex, actualType.getName(), expectedType.getName());
            throw new SpecificationViolationException(msg, method);
          }
        }
        indexed++;
      } else {
        if (indexed > 0) {
          throw new SpecificationViolationException("You're not allowed to mix @Named and indexed params", method);
        }
        final String name = named.value();
        if (metadata != null) {
          final NamedParameterDescriptor descriptor;
          try {
            descriptor = metadata.getNamedParameterDescriptor(name);
          } catch (final QueryParameterException e) {
            throw new SpecificationViolationException(String.format("Query doesn't have named parameter '%s'", name), method, e);
          }
          final Type expectedType = descriptor.getExpectedType();
          // TODO: warn about unknown expected type? The only case I know when this happens is SQL query.
          if (expectedType != null) {
            final Class<?> expectedClass = expectedType.getReturnedClass();
            final Class<?> actualClass = getActualParameterType(type, true);
            if (!Collection.class.isAssignableFrom(actualClass) && !expectedClass.isAssignableFrom(actualClass)) {
              final String msg = String.format("Named parameter '%s' of type %s cannot be assigned to query parameter of type %s", name, actualClass.getName(), expectedClass.getName());
              throw new SpecificationViolationException(msg, method);
            }
          }
        }
        if (!names.add(name)) {
          throw new SpecificationViolationException("Duplicate named parameters '" + name + "'", method);
        }
      }
    }
  }

  @NotNull
  private Class<?> getActualParameterType(@NotNull final Class<?> declaredType, final boolean handleArrays) {
    final QueryArgumentTransformer transformer = transformers.get(declaredType);
    final Class<?> transformedType = transformer == null ? declaredType : transformer.getTargetType();
    // TODO: check collection params too
    final Class<?> arrayUnwrappedType = handleArrays && transformedType.isArray() ? transformedType.getComponentType() : transformedType;
    return primitive(arrayUnwrappedType);
  }

  /**
   * Copied from commons-binutils (org.apache.commons.beanutils.converters.AbstractConverter).
   * <p/>
   * Change primitive Class types to the associated wrapper class.
   * @param type The class type to check.
   * @return The converted type.
   */
  @SuppressWarnings({ "IfStatementWithTooManyBranches", "ObjectEquality" })
  @NotNull
  private static Class<?> primitive(@NotNull final Class<?> type) {
    if (!type.isPrimitive()) {
      return type;
    } else if (type == Integer.TYPE) {
      return Integer.class;
    } else if (type == Double.TYPE) {
      return Double.class;
    } else if (type == Long.TYPE) {
      return Long.class;
    } else if (type == Boolean.TYPE) {
      return Boolean.class;
    } else if (type == Float.TYPE) {
      return Float.class;
    } else if (type == Short.TYPE) {
      return Short.class;
    } else if (type == Byte.TYPE) {
      return Byte.class;
    } else if (type == Character.TYPE) {
      return Character.class;
    } else {
      return type;
    }
  }

  private void limitCheck(final Class<?> type) {
    limitOffsetCheck(limit, singleFind, type, method);
    limit = true;
  }

  private void offsetCheck(final Class<?> type) {
    limitOffsetCheck(offset, singleFind, type, method);
    offset = true;
  }

  private void limitOffsetCheck(final boolean value,
                                final boolean singleFind,
                                final Class<?> type,
                                final Method m) {
    if (value) {
      throw new SpecificationViolationException("@Limit/@Offset can appear only on one parameter", m);
    } else if (singleFind) {
      throw new SpecificationViolationException("@Limit/@Offset isn't allowed on single-object finders", m);
    }
    if (type != Integer.class && type != Integer.TYPE) {
      throw new SpecificationViolationException(
          "@Limit and @Offset argument type must be either int or java.lang.Integer", m);
    }
  }

  public Set<String> getNamedParameterNames() {
    return Collections.unmodifiableSet(names);
  }

  public int getIndexedParametersCount() {
    return indexed;
  }
}
